Address subclass: #ExternalReplaceAddress
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Actor-Ext'!
ExternalReplaceAddress comment:
'Class ReplaceActor provides behavior replacement messages.
This allows behavior replacement messages to be computed externally and sent to the actor.
This is useful when the replacement behavior needs some non local information to be computed and therefore has to be computed externally by another actor.
Meanwhile the actor becomes insensitive (because it cannot accept any message with no behavior) while waiting (passively) for its replacement behavior.

Class ReplaceActor is a subclass of kernel class Actor.
Method replace: directly handles behavior replacement messages at the actor level (and not at the behavior level). This is necessary because while insensitive the actor may not accept any message!!

Example of use is the management of overdraft. See class ProtectedBankAccount in category Actalk-Ext-Replace-Ext-Ex.
Also immediate extensions are implementations of futures, see class FutureActor.'!


!ExternalReplaceAddress methodsFor: 'behavior replacement'!

replace: replacementBehavior
	"Specify behavior replacement (assignment of a new active object and activity to the address).
	Note that this replacement behavior message is processed by the address
	and consequently may be sent externally by another actor.
	(See example of class OverdraftManager in category Actalk-Ext-Actor-Ext-Ex.)"

	self startActivity: (replacementBehavior activityClass new activeObject: replacementBehavior)! !

ExternalReplaceAddress subclass: #FutureActor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Actor-Ext'!
FutureActor comment:
'Class FutureActor implements futures (that is objects/actors whose behavior/value has not been assigned yet) simply as a newly created actor with no behavior yet.
In other words a FutureActor is simply a born insensitive actor.
It will buffer all messages until its real nature (behavior) gets computed (externally and assigned through a replacement behavior message).

Actually there is no need for any implementation. No instance variable nor method is defined!!
Its superclass ReplaceActor could be used as well. But we define class FutureActor as a way to clearly identify the use of futures.

Note that the limitation of this model of future is that it value when computed may only be defined through a behavior.
Because Actalk is not fully uniform (not everything is an actor), basic entities such as numbers cannot be easily defined through behaviors. A future actor cannot become such primitive objects.
Future actors are useful to implement external actors, that is, in the terminology of Gul Agha, actors which provide interfacing with other program modules. We may provide the actual connection at execution time, meanwhile future actors will buffer all messages.

One alternative to this trivial implementation relies on more traditional objects and assignment, as well as semaphores. This could be found in category Actalk-Ext-Abcl and will be used to implement eager type (also called future type) message passing.'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

FutureActor class
	instanceVariableNames: ''!


!FutureActor class methodsFor: 'example'!

example
	"self example"

	self new
		deposit: 300 replyTo: Print;
		replace: (BankAccount new balance: 200);
		withdraw: 400 replyTo: Print;
		balanceAndReplyTo: Print! !

EnabledSelectorsActivity subclass: #SingleMessageEnabledSelectorsActorActivity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Actor-Ext'!
SingleMessageEnabledSelectorsActorActivity comment:
'Class AcceptReplaceActorBehavior is a subclass of class ReplaceActorBehavior and provides control over acceptance of messages.
Basically it looks only for messages that it knows (selectors of defined methods). This is useful to disable method acceptance (without relying on trapping errors and resending messages as in class NullBoundedBuffer).
Method nextMessage is redefined to look for messages that the class knows (defines or inherits).
See class AcceptNullBoundedBuffer in category Actalk-Ext-Replace-Ext-Ex as the resulting improved version of the bounded buffer example.'!


!SingleMessageEnabledSelectorsActorActivity methodsFor: 'initialize'!

privateInitialize
	"Initialize the collection of enabled selectors as the set of all script selectors."

	super privateInitialize.
	enabledSelectors := oself class allScriptSelectors! !

!SingleMessageEnabledSelectorsActorActivity methodsFor: 'activity setting'!

body
	"Accept a single message."

	self acceptNextMessage! !

!SingleMessageEnabledSelectorsActorActivity methodsFor: 'message handling'!

performMessage: aMessage
	"Return to basic way of performing message to bypass the implicit reply value handling."

	^oself performMessage: aMessage! !

!SingleMessageEnabledSelectorsActorActivity methodsFor: 'default classes'!

addressClass
	"This is to bypass the ImplicitReplyAddress inherited default class
	because we actually don't need to use the implicit reply feature,
	we just want to inherit EnabledSelectorsActivity features."

	^Address! !
