GenericSendAddress subclass: #ImplicitReplyAddress
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ImplicitReply'!
ImplicitReplyAddress comment:
'Classes IRActor and IRActorBehavior (IR stands for Implicit Reply) implement a better integration of message passing between actors within standard Smalltalk framework.
With use of implicit reply actors the standard Smalltalk way of writing and using methods is preserved. Implicit reply and returned value is performed by using the ^. One does need to add arguments to methods to convey the extra reply destination argument because it becomes implicit.
Sending a message to an IRActor returns a future object which will eventually contain the value returned by the method performed. By deciding not to use the return future, or to wait for its value, one simulates Abcl asynchronous (past) or synchronous (now) types of message passing.'!


!ImplicitReplyAddress methodsFor: 'message passing'!

insertReplyDestination: r intoMessage: aMessage
	"Insert the reply destination as a new implicit variable of the (extended) message.
	Return the extended message."

	^(ReplyToMessage new
			setSelector: aMessage selector arguments: aMessage arguments;
			replyTo: r)!

receiveMessage: aMessage withSendType: type inMessageQueue: queue
	"Recognize asynchronous type."

	type = #future
		ifTrue: [^self futureSend: aMessage withFutureClass: SAFuture inMessageQueue: queue].
	^super receiveMessage: aMessage withSendType: type inMessageQueue: queue!

sendTypeOf: aMessage
	"Assume that every message sent should create a future to hold the reply.
	The future is returned (as for a future type message passing)."

	^#future! !

!ImplicitReplyAddress methodsFor: 'default classes'!

activityClass
	"ImplicitReplyAddress way of receiving a future type message
	makes assumption that the future object will eventually receive its value
	computed by
	ImplicitReplyActivity method
		performMessage:"

	^ImplicitReplyActivity! !

!ImplicitReplyAddress methodsFor: 'compatibility constraints'!

activityConstraint
	"ImplicitReplyAddress way of receiving a future type message
	makes assumption that the future object will eventually receive its value
	computed by
	ImplicitReplyActivity method
		performMessage:"

	^ImplicitReplyActivity! !

ImplicitReplyAddress subclass: #OptimizedImplicitReplyAddress
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ImplicitReply'!
OptimizedImplicitReplyAddress comment:
'Classes OIRActor and OIRActorBehavior (O stands for Optimized and IR stands for Implicit Reply) implement an optimized version of implicit reply (respective superclasses IRActor and IRActorBehavior).
The point is following: many transmissions are asynchronous. In such a case creating a future means a waste of resources.

A simple way of tagging asynchronous message passing is by using special selectors (starting with capital letter A). In that case, the message sending and computation proceeds as in standard actor case.

Note that this means that a same selector may not be used both for asynchronous message passing and other types (future or/and synchronous). However this is rarely the case, except when the value is of no use but the sender wants to be able to synchronize on the acknowledgement or not. In such a case the method may be duplicated.'!


!OptimizedImplicitReplyAddress methodsFor: 'message passing'!

sendTypeOf: aMessage
	"Dispatch onto asynchronous or future type message passing,
	according to the first letter of the selector (A for asynchronous)."

	^(aMessage selector at: 1) = $A
		ifTrue: [#asynchronous]
		ifFalse: [#future]! !

Activity subclass: #ImplicitReplyActivity
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ImplicitReply'!


!ImplicitReplyActivity methodsFor: 'message handling'!

performMessage: aMessage
	"Perform the message and assign (send) the returned value to the future (reply destination)."

	aMessage replyTo reply: (super performMessage: aMessage)! !

!ImplicitReplyActivity methodsFor: 'default classes'!

addressClass
	"ImplicitReplyActivity method
		performMessage:
			makes assumption that the message has been extended
			(into a ReplyMessage conveying a future object implicit reply destination) by
	ImplicitReplyAddress method
		insertReplyDestination:intoMessage:"

	^ImplicitReplyAddress! !

!ImplicitReplyActivity methodsFor: 'compatibility constraints'!

addressConstraint
	"Enforce assumption of method addressClass"

	^ImplicitReplyAddress! !

Message subclass: #ReplyToMessage
	instanceVariableNames: 'replyTo '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ImplicitReply'!
ReplyToMessage comment:
'Class ReplyToMessage is a subclass of class Message. It adds an extra instance variable to implicitly contain and convey the reply destination. This class is used for implicit reply actors (classes IRActor and IRActorBehavior).

Instance Variables:

	replyTo		<Future>	the implicit reply destination (a future object) conveyed within the message.'!


!ReplyToMessage methodsFor: 'accessing'!

replyTo
	^replyTo!

replyTo: replyDestination
	replyTo := replyDestination! !

ActiveObject subclass: #ImplicitReplyObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ImplicitReply'!


!ImplicitReplyObject methodsFor: 'default classes'!

activityClass
	"This active object class only purpose is to dispatch onto this activity class."

	^ImplicitReplyActivity! !

ImplicitReplyObject subclass: #OptimizedImplicitReplyObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-ImplicitReply'!


!OptimizedImplicitReplyObject methodsFor: 'default classes'!

addressClass
	"This active object class only purpose is to dispatch onto this address class."

	^OptimizedImplicitReplyAddress! !
