ConcurrentActivity subclass: #SuspendActivity
	instanceVariableNames: 'suspendedSubProcessAssociationList controlSemaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Suspend'!
SuspendActivity comment:
'Class SuspendActorBehavior implements suspension of methods.
Within a method body, one may express suspension on some condition.
This is analog to ConcurrentSmalltalk-II relinquish: method.
This is also analog to Kessel monitors except that monitors synchronize external processes, as here processes are internally created for every starting request.

We need to be able to suspend some method computation, and then start another one or restart one previously suspended (and whose suspension condition became false).
Therefore we need a main control behavior process. It will act as a scheduler of message/method activations. It launchs suprocesses for computing requests. It gets then suspended but gets resumed once a method computation is suspended or completed.
The main control behavior process is following.
It first checks if one previously suspended method subprocess is ready to resume (by checking their conditions). If the case, it resumes it. Otherwise it creates a new behavior subprocess to process next pending message. (This is a lazy creation as it first waits for a pending message if needed). Then (in both cases) it suspends until next method completion or suspension.

Class SuspendActorBehavior is defined as a subclass of class ConcurrentActorBehavior as it holds subprocesses.

Instance Variables:

	suspendedProcessList 	<OrderedCollection>	holds the ordered collection of waiting (suspended) subprocesses.
							Each item of the collection is an association:
								<condition block , semaphore on which the process waits>.

	controlSemaphore		<Semaphore>	is the main control semaphore to synchronize the main central behavior process
							and the starting/suspension/resumption/completion of subprocesses created to compute messages/methods.'!


!SuspendActivity methodsFor: 'initialize'!

privateInitialize
	super privateInitialize.
	suspendedSubProcessAssociationList := OrderedCollection new.
	controlSemaphore := Semaphore new! !

!SuspendActivity methodsFor: 'activity setting'!

body
	"The behavior body specifies the main control body behavior.
	If there is a suspended message whose resumption condition has become true, resume it.
	Otherwise create a new subprocess for next pending message.
	Then current main activity process gets suspended (until next method completion or suspension)."

	| association |
	[true] whileTrue:
		["First look for (the first) suspended subprocess whose resumption condition has become true."
		association := self lookForFirstResumeableSubProcess.
		association isNil
			ifFalse:		"If found, then resume it."
				[self resumeSuspendedSubProcessAssociation: association]
			ifTrue:		"Otherwise, start a new subprocess to accept next message."
				[self startSubProcessFor: self nextMessage].
		"In both cases, wait for resumption (after next method completion or suspension)."
		controlSemaphore wait]! !

!SuspendActivity methodsFor: 'suspension management'!

addSuspendedSubProcessOnCondition: resumptionConditionBlock andSemaphore: aSemaphore
	"association = <key: resumption condition block , value: semaphore on which the process waits>."

	suspendedSubProcessAssociationList addLast:
		(Association new key: resumptionConditionBlock value: aSemaphore)!

lookForFirstResumeableSubProcess
	"Look for the first suspended subprocess whose resumption condition has become true.
	If one is found, return the corresponding association, otherwise return nil."

	"association = <key: resumption condition block , value: semaphore on which the process waits>."

	^suspendedSubProcessAssociationList isEmpty
		ifTrue:
			[nil]
		ifFalse:
			[suspendedSubProcessAssociationList	
				detect: [:association | association key value]
				ifNone: [nil]]!

resumeSuspendedSubProcessAssociation: association
	"Resume a suspended subprocess by removing the association from the list and signaling the semaphore on which it is waiting."

	"association = <key: resumption condition block , value: semaphore on which the process waits>."

	suspendedSubProcessAssociationList remove: association.
	association value signal!

subProcessWaitUntil: resumptionConditionBlock
	"Suspend current method computation on a resumption condition (block)."
	"Equivalent to relinquish: construct in ConcurrentSmalltalk-II."

	| aSemaphore |
	aSemaphore := Semaphore new.
	self addSuspendedSubProcessOnCondition: resumptionConditionBlock andSemaphore: aSemaphore.
	"First resume the main activity body process,"
	controlSemaphore signal.
	"Then suspend current subprocess onto the semaphore."
	aSemaphore wait

	"Note that when signaled, the suspended subprocess resumes here,
	and immediately returns to resumption of processing the method body which had been suspended."! !

!SuspendActivity methodsFor: 'events'!

kernelEventComplete: aMessage
	"Resume the main control behavior process after completing activation of a message."

	super kernelEventComplete: aMessage.
	controlSemaphore signal! !

!SuspendActivity methodsFor: 'process control'!

terminate
	"Terminate all suspended subprocesses at first."

	"association = <key: resumption condition block , value: semaphore on which the process waits>."
	suspendedProcessAssociationList do:
		"For each suspension semaphore, terminate the waiting process."
		[:association | association value terminateProcess].
	super terminate! !

ImplicitReplyObject subclass: #SuspendObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Suspend'!


!SuspendObject methodsFor: 'waitUntil construct'!

waitUntil: resumptionConditionBlock
	"Suspend current method computation on a resumption condition (block)."
	"Equivalent to relinquish: construct in ConcurrentSmalltalk-II."

	"If resumption condition is true do nothing."
	resumptionConditionBlock value ifFalse:
		[self activity subProcessWaitUntil: resumptionConditionBlock]! !

!SuspendObject methodsFor: 'default classes'!

activityClass
	"SuspendObject method
		waitUntil:
			may call
	SuspendActivity method
		subProcessWaitUntil:."

	^SuspendActivity! !

!SuspendObject methodsFor: 'compatibility constraints'!

activityConstraint
	"Enforce constraint defined by method activityClass."

	^SuspendActivity! !
